module(..., package.seeall)

--for reporting
local report = require("reporter.report")
recalc = require("Association_recalc")

local json = require("reporter.dkjson")
local openProject = require ("openProject")

--require for java support
require "java"
require "lfs"


function select_ontology_path(start_path)
	return tda.BrowseForFile("Open", "OWL File (*.owl;*.xml;*.rdf)\nAll Files (*)", start_path or "", "", false)
end

function select_and_load_ontology(start_path)
	print ("start load")
	local path = select_ontology_path(start_path or "")
	if path ~= "" then
		load_ontology(path)
		openProject.closeForm()
		recalc.associations()
		return true
	end
	--openProject.closeForm()
	return false
end

function str_starts_with(str, start)
   return string.sub(str, 1, string.len(start)) == start
end

local function file_size_in_kb(file_path)
	return math.floor((lfs.attributes(file_path, "size") or 0) / 1024)
	--return 1024
end

local function file_name(file_path)
	local name = string.gsub(file_path, '.*\\', '')
	return name
end

function disabled_renderers(render_pefs)
	local disable_renderer_names = {}
	for renderer_name, status in pairs(render_pefs) do
		if status == false or status == "false" then
			table.insert(disable_renderer_names, renderer_name)
		end
	end
	return disable_renderer_names
end

function load_ontology(ontology_path)
--original, now not needed
--	local status, render_prefs_table = pcall(parameters.createParametersTable)
	render_prefs_table = status and render_prefs_table or {}
	-- print(dumptable(render_prefs_table))


	-- Log "ontology load" event.
	report.event("ViziQuer_LoadOntology", {
		file_size_in_kb = file_size_in_kb(ontology_path),
		file_name = file_name(ontology_path),
		disabled_renderers = disabled_renderers(render_prefs_table),
		render_preferences_set = status,
	})

--[[original
	local java_class_name = "OWLGrEdImport"
	local method_name = "read"
]]
	local java_class_name = "readowl.ReadOWL"
	local class_method_name = "readClass"
	local assoc_method_name = "readAssoc"
	local attribute_method_name = "readAttribute"
	local subClass_method_name = "readSubClass"

--clear ontology
		lQuery("Onto#Element"):each(
		function(obj)
			obj:delete()
		end)

	--ont_as_table_str = java.call_static_class_method(java_class_name, method_name, serialized_args)
	classSpy(java_class_name, class_method_name, ontology_path)
	AssocSpy(java_class_name, assoc_method_name, ontology_path)
	AttributeSpy(java_class_name, attribute_method_name, ontology_path)
	SubClassSpy(java_class_name, subClass_method_name, ontology_path)
	AddOwlThing()
end
--[[
function import_from_owlapi(table_as_string)
  tda.CallFunctionWithPleaseWaitWindow("loadOntology.import_from_owlapi_helper", table_as_string)
end
--]]

function classSpy(java_class_name, method_name, ontology_path)
	local classTableString = java.call_static_class_method(java_class_name, method_name, ontology_path)
	if str_starts_with(classTableString, "error - ") then
		report.event("ViziQuer_owlapi_error_class", {message = classTableString})
		error(classTableString)
	else
		local table_representation = json.decode(classTableString)
		parseClasses(table_representation)
	end
end

function parseClasses (classTable)
	for i,klasesInfo in pairs(classTable) do
		klase = lQuery.create("Onto#Class")
		klase:attr({type = "Ontology", localname = klasesInfo.name, URI = klasesInfo.URI, namespace = klasesInfo.namespace, name = klasesInfo.name})
	end
end

function AssocSpy(java_class_name, method_name, ontology_path)
	local AssocTableString = java.call_static_class_method(java_class_name, method_name, ontology_path)
	if str_starts_with(AssocTableString, "error - ") then
		report.event("ViziQuer_owlapi_error_assoc", {message = AssocTableString})
		error(AssocTableString)
	else
		local table_representation = json.decode(AssocTableString)
		--print (dumptable(table_representation))
		parseAssoc(table_representation)
	end
end

function parseAssoc (AssocTable)
	for i,assocInfo in pairs(AssocTable) do
		assoc = lQuery.create("Onto#Association")
		assoc:attr({type = "Ontology", localname = assocInfo.name, URI = assocInfo.URI, namespace = assocInfo.namespace, name = assocInfo.name})
		for i,assocRange in pairs(assocInfo.range) do
			rangeKlase = lQuery("Onto#Class[URI = '".. assocRange .. "']")
			assoc:link ("range", rangeKlase)
		end
		for i,assocDomain in pairs(assocInfo.domain) do
			domainKlase = lQuery("Onto#Class[URI = '".. assocDomain .. "']")
			assoc:link ("domain", domainKlase)
		end
	end
end

function AttributeSpy(java_class_name, method_name, ontology_path)
	local AttributeTableString = java.call_static_class_method(java_class_name, method_name, ontology_path)
	if str_starts_with(AttributeTableString, "error - ") then
		report.event("ViziQuer_owlapi_error_attribute", {message = AttributeTableString})
		error(AttributeTableString)
	else
		print ("attribute string")
		print (AttributeTableString)
		local table_representation = json.decode(AttributeTableString)
		--print (dumptable(table_representation))
		parseAttribute(table_representation)
	end
end

function parseAttribute (AttributeTable)
	print(dumptable(AttributeTable))
	--if AttributeTable ~= nil then
		for i,attributeInfo in pairs(AttributeTable) do
			attribute = lQuery.create("Onto#Attribute")
			attribute:attr({localname = attributeInfo.name, URI = attributeInfo.URI, namespace = attributeInfo.namespace, name = attributeInfo.name})
			for i,attributeDomain in pairs(attributeInfo.domain) do
				domainKlase = lQuery("Onto#Class[URI = '".. attributeDomain .. "']")
				attribute:link ("class", domainKlase)
			end
		end
	--end
end

function SubClassSpy(java_class_name, method_name, ontology_path)
	local SubClassTableString = java.call_static_class_method(java_class_name, method_name, ontology_path)
	if str_starts_with(SubClassTableString, "error - ") then
		report.event("ViziQuer_owlapi_error_subClass", {message = SubClassTableString})
		error(SubClassTableString)
	else
		local table_representation = json.decode(SubClassTableString)
		print (dumptable(table_representation))
		parseSubClass(table_representation)
	end
end

function parseSubClass (SubClassTable)
	for i,classInfo in pairs(SubClassTable) do
		--print (classInfo.URI)
		local klase = lQuery("Onto#Class[URI = '".. classInfo.URI .. "']")
		for i,subClassInfo in pairs(classInfo.subClass) do
			print (subClassInfo)
			local subKlase = lQuery("Onto#Class[URI = '".. subClassInfo .. "']")
			klase:link ("subClass", subKlase)
		end
	end
end

function AddOwlThing()
	local thing = lQuery.create("Onto#Class", {value = value})
	thing:attr({type = "Ontology", localname = "Thing", URI = "http://www.w3.org/2002/07/owl#Thing", namespace = "http://www.w3.org/2002/07/owl#", name = "Thing"})
	lQuery("Onto#Class:not(/superClass)"):each(
			function (class)
				class:link("superClass", thing)
			end
		)
	thing:remove_link("superClass", thing)

end

function eval(string_to_eval)
  log("---> eval called, got length", #string_to_eval)
  return loadstring("return " .. string_to_eval)()
end
--log = function () end
