module(..., package.seeall)

require "core"
require "lpeg"
require "re"

function sparqlExpressionGrammar()
	local grammar = [[
		Main <- (Expression space !.)
		Expression <- ({:ConditionalOrExpression: ConditionalOrExpression :}) -> {}
		ConditionalOrExpression <- 	(ConditionalAndExpression  ( space '||' space ConditionalAndExpression )*) -> {}
		ConditionalAndExpression <- ({:ConditionalAndExpression: ValueLogicalA :}) -> {}
		ValueLogicalA <- (ValueLogical (space '&&' space ValueLogical )*) -> {}
		ValueLogical <- ({:RelationalExpression: RelationalExpression :}) -> {}
		RelationalExpression <- ({:NumericExpression: NumericExpression:} ((space {:Relation: '=' :} space NumericExpression )
				/ (space {:Relation: '!=' :} space NumericExpression )
				/ (space {:Relation: '<' :} space NumericExpression )
				/ (space {:Relation: '>'  :} space NumericExpression )
				/ (space {:Relation: '<=' :} space NumericExpression )
				/ (space {:Relation: '>=' :} space NumericExpression )
				/ (space {:Relation: 'IN' :} space ExpressionList )
				/ (space {:Relation: 'NOT' space 'IN' :} space ExpressionList ))?) ->{}
		NumericExpression <- ({:AdditiveExpression: AdditiveExpression:}) -> {}
		AdditiveExpression <- 	({:MultiplicativeExpression: MultiplicativeExpression:} ( MultiplicativeExpressionList )*) -> {}
		MultiplicativeExpressionList <- ((space {:Additive: '+' :} space {:MultiplicativeExpression: MultiplicativeExpression:}) 
									/ (space {:Additive: '-' :} space {:MultiplicativeExpression: MultiplicativeExpression:})
									/ ( NumericLiteralPositive / NumericLiteralNegative ) 
									( (space '*'space UnaryExpression ) / (space '/' space UnaryExpression ) )*) -> {}
		MultiplicativeExpression <- ({:UnaryExpression: UnaryExpression:} ( UnaryExpressionList)*) -> {}
		UnaryExpressionList <- ((space {:Unary: '*' :}space  {:UnaryExpression: UnaryExpression:} )
							/ (space {:Unary: '/' :} space {:UnaryExpression: UnaryExpression:}) ) -> {}
		UnaryExpression <- ((space {:Additive: '!' :} space {:PrimaryExpression: PrimaryExpression:}) 
							/(space {:Additive: '+' :} space {:PrimaryExpression: PrimaryExpression:})
							/(space {:Additive: '-' :} space {:PrimaryExpression: PrimaryExpression:})
							/ {:PrimaryExpression: PrimaryExpression:}) -> {}
		PrimaryExpression <- BooleanLiteral / BuiltInCall / RDFLiteral / BrackettedExpression / iriOrFunction  / NumericLiteral /  Var / QName/ LName 
		PrimaryExpression2 <- BooleanLiteral / BuiltInCall / RDFLiteral / BrackettedExpression / iriOrFunction  / NumericLiteral /  Var / LName 
		LName <- ({:LName: Chars_String:}) -> {}
		QName <- ({:Path: path :} PrimaryExpression2) -> {}
		path <- ((({"INV(" Chars_String ")"} / {"^"? Chars_String} ) ".")*) -> {}
		BrackettedExpression <- ('(' space {:BrackettedExpression: Expression :} space ')') -> {}
		BuiltInCall <-  ({:Aggregate: Aggregate :} / {:FunctionExpression: FunctionExpression :} / {:RegexExpression: RegexExpression :} / {:SubstringExpression: SubstringExpression :} / {:StrReplaceExpression: StrReplaceExpression :} / ExistsFunc / NotExistsFunc) -> {}
		FunctionExpression <- ('STR' '(' space {:Expression: Expression :} space ')'
			/ ({:Function: 'LANG':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'LANGMATCHES':} '(' space Expression ',' space Expression space ')')
			/ ({:Function: 'DATATYPE':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'BOUND':} '(' space Var space ')')
			/ ({:Function: 'IRI':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'URI':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'BNODE':} ( '(' space {:Expression: Expression :} space ')' / NIL ))
			/ ({:Function: 'RAND':} NIL)
			/ ({:Function: 'ABS':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'CEIL':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'FLOOR':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'ROUND':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'CONCAT':} ExpressionList)
			/ ({:Function: 'STRLEN':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'UCASE':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'LCASE':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'ENCODE_FOR_URI':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'CONTAINS':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'STRSTARTS':} '(' space Expression space  ',' space Expression space ')')
			/ ({:Function: 'STRENDS':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'STRBEFORE':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'STRAFTER':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'YEAR':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'MONTH':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'DAY':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'HOURS':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'MINUTES':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'SECONDS':} '(' {:Expression: Expression :} space ')')
			/ ({:Function: 'TIMEZONE':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'TZ':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'NOW':} NIL)
			/ ({:Function: 'UUID':} NIL)
			/ ({:Function: 'STRUUID':} NIL)
			/ ({:Function: 'MD5':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'SHA1':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'SHA256':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'SHA384':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'SHA512':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'COALESCE':} ExpressionList)
			/ ({:Function: 'IF':} '(' space Expression space ',' space Expression ',' space Expression space ')')
			/ ({:Function: 'STRLANG':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'STRDT':} '(' space Expression space ','space  Expression space ')')
			/ ({:Function: 'sameTerm':} '(' space Expression space ',' space Expression space ')')
			/ ({:Function: 'isIRI':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'isURI':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'isBLANK':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'isLITERAL':} '(' space {:Expression: Expression :} space ')')
			/ ({:Function: 'isNUMERIC':} '(' space {:Expression: Expression :} space ')')) -> {}
		iriOrFunction <- ({:iri: iri:} {:ArgList: ArgList?:}) -> {}
		ArgList <- ('(' space {:DISTINCT: 'DISTINCT'? :} space {:ArgListExpression: ArgListExpression :} space ')' / NIL ) -> {}
		ArgListExpression <-  (Expression ( ',' space Expression )*) -> {}
		RDFLiteral <- 	({:RDFLiteral: RDFLiteralA:}) -> {}
		RDFLiteralA <- 	({String} ( {:LANGTAG: LANGTAG:} / ( '^^' {:iri: iri:} ) )?) -> {}
		NumericLiteral <-	({:NumericLiteral: NumericLiteralUnsigned / NumericLiteralPositive / NumericLiteralNegative :})->{}
		NumericLiteralUnsigned <- INTEGER / DECIMAL / DOUBLE
		NumericLiteralPositive <- INTEGER_POSITIVE / DECIMAL_POSITIVE / DOUBLE_POSITIVE
		NumericLiteralNegative <- INTEGER_NEGATIVE / DECIMAL_NEGATIVE / DOUBLE_NEGATIVE
		BooleanLiteral <- ({:BooleanLiteral: 'true' / 'false':}) -> {}
		Var <- ({:Var: VAR1 / VAR2 :}) ->{}
		VAR1 <- {'?' VARNAME}
		VAR2 <- {'$' VARNAME}
		-- VARNAME <- ({([A-Za-z] / "_") ([A-Za-z] / "_" / [0-9])*})
		ExpressionList <- (NIL / '(' space Expression space  ( ',' space Expression )* space ')' )
		RegexExpression <- ('REGEX' '(' space Expression space  ',' space Expression ( ',' space Expression )? space ')') -> {}
		SubstringExpression <- ('SUBSTR' '(' space Expression space  ',' space Expression ( ',' space Expression )? space ')') -> {}
		StrReplaceExpression <- ('REPLACE' '(' space Expression space  ',' space Expression ',' space Expression ( ',' space Expression )? space ')') ->{}
		ExistsFunc <- ({:ExistsFunc: 'EXISTS' space GroupGraphPattern :}) -> {}
		NotExistsFunc <- ({:NotExistsFunc: 'NOT' space  'EXISTS' space GroupGraphPattern :}) -> {}
		GroupGraphPattern <- ( {:PrimaryExpression: PrimaryExpression :}) -> {}
		Aggregate <- (({:Aggregate: 'COUNT' :} '(' space {:DISTINCT: 'DISTINCT'? :} space ( {'*'} / {:Expression: Expression :} ) space ')')
					/ ({:Aggregate: 'SUM' :} '(' space {:DISTINCT: 'DISTINCT'? :} space {:Expression: Expression :} space ')')
					/ ({:Aggregate: 'MIN' :} '(' space {:DISTINCT: 'DISTINCT'? :} space {:Expression: Expression :} space ')')
					/ ({:Aggregate: 'MAX' :} '(' space {:DISTINCT: 'DISTINCT'? :} space{:Expression: Expression :} space ')')
					/ ({:Aggregate: 'AVG' :} '(' space {:DISTINCT: 'DISTINCT'? :} space {:Expression: Expression :} space ')')
					/ ({:Aggregate: 'SAMPLE' :} '(' space {:DISTINCT: 'DISTINCT'? :} space{:Expression: Expression :} space ')')
					/ ({:Aggregate: 'GROUP_CONCAT' :} '(' space {:DISTINCT: 'DISTINCT'? :} space{:Expression: Expression :} ( ';' space 'SEPARATOR' '=' {:SEPARATOR: String :} )? space ')')) -> {}
		iri <- ({:IRIREF: IRIREF:} / {:PrefixedName: PrefixedName:}) -> {}
		PrefixedName <- {PNAME_LN} / {PNAME_NS}
		IRIREF <- '<' ([A-Za-z] / "_" / [0-9])* '>'
		-- IRIREF <- '<' ([^<>"{}|^`\]-[#x00-#x20])* '>'
		PNAME_NS <- PN_PREFIX? ':'
		PNAME_LN <- PNAME_NS PN_LOCAL
		String <- STRING_LITERAL1 / STRING_LITERAL2 / STRING_LITERAL_LONG1 / STRING_LITERAL_LONG2
		-- STRING_LITERAL1 <- "'" ( ([^#x27#x5C#xA#xD]) / ECHAR )* "'"
		STRING_LITERAL1 <- "'" ( ([A-Za-z] / [-_.:] / [0-9]) / ECHAR )* "'"
		-- STRING_LITERAL2 <- '"' ( ([^#x22#x5C#xA#xD]) / ECHAR )* '"'
		STRING_LITERAL2 <- '"' (  ([A-Za-z] / [-_.:] / [0-9]) / ECHAR )* '"'
		STRING_LITERAL_LONG1 <- "'''" ( ( "'" / "''" )? ( [^'\] / ECHAR ) )* "'''"
		STRING_LITERAL_LONG2 <- '"""' ( ( '"' / '""' )? ( [^"\] / ECHAR ) )* '"""'
		ECHAR <- '\' [tbnrf\"']
		NIL <- '('  ')'
		WS <- ({([A-Za-z] / "_") ([A-Za-z] / "_" / [0-9])*})
		--WS <- #x20 / #x9 / #xD / #xA
		LANGTAG <- ({'@' [A-Za-z] ([A-Za-z])* ('-' [a-zA-Z0-9]+)*})
		INTEGER	  <-  	[0-9]+
    	DECIMAL	  <-  	[0-9]* '.' [0-9]+
		DOUBLE	  <-  	([0-9]+ '.' [0-9]* EXPONENT) / ('.' ([0-9])+ EXPONENT) / (([0-9])+ EXPONENT)
		INTEGER_POSITIVE	  <-  	'+' INTEGER
		DECIMAL_POSITIVE	  <-  	'+' DECIMAL
		DOUBLE_POSITIVE	  <- 	'-' INTEGER
		INTEGER_NEGATIVE	  <-  	'-' INTEGER
		DECIMAL_NEGATIVE	  <-  	'-' DECIMAL
		DOUBLE_NEGATIVE	  <- 	'-' DOUBLE
		EXPONENT	  <-  	[eE] [+-]? [0-9]+
		-- PN_CHARS_BASE <- [A-Z] / [a-z] / [#x00C0-#x00D6] / [#x00D8-#x00F6] / [#x00F8-#x02FF] / [#x0370-#x037D] / [#x037F-#x1FFF] / [#x200C-#x200D] / [#x2070-#x218F] / [#x2C00-#x2FEF] / [#x3001-#xD7FF] / [#xF900-#xFDCF] / [#xFDF0-#xFFFD] / [#x10000-#xEFFFF]
		PN_CHARS_BASE <- ()[A-Z] / [a-z] 
		PN_CHARS_U <- PN_CHARS_BASE / '_'
		-- VARNAME	 <- 	( PN_CHARS_U / [0-9] ) ( PN_CHARS_U / [0-9] / #x00B7 / [#x0300-#x036F] / [#x203F-#x2040] )*
		VARNAME	 <- 	( PN_CHARS_U / [0-9] ) ( PN_CHARS_U / [0-9] /  [A-Z] / [a-z])*
		-- PN_CHARS <- 	PN_CHARS_U / '-' / [0-9] / #x00B7 / [#x0300-#x036F] / [#x203F-#x2040]
		PN_CHARS <- 	PN_CHARS_U / '-' / [0-9] / [A-Z] / [a-z]
		-- PN_PREFIX <- PN_CHARS_BASE ((PN_CHARS/'.')* PN_CHARS)?
		PN_PREFIX <- Chars_String
		-- PN_LOCAL <- (PN_CHARS_U / ':' / [0-9] / PLX ) ((PN_CHARS / '.' / ':' / PLX)* (PN_CHARS / ':' / PLX) )?
		PN_LOCAL <- Chars_String
		PLX	 <- PERCENT / PN_LOCAL_ESC
		PERCENT	 <- '%' HEX HEX
		HEX <- [0-9] / [A-F] / [a-f]
		PN_LOCAL_ESC <- '\' ( '_' / '~' / '.' / '-' / '!' / '$' / '&' / "'" / '(' / ')' / '*' / '+' / ',' / ';' / '=' / '/' / '?' / '#' / '@' / '%' )
		Chars_String <- (([A-Za-z] / "_") ([A-Za-z] / "_" / [0-9])*)
		space <- ((" ")*)
	]]
	return grammar
end

function parsesparqlExpressionGrammar(str)
	local gramar = [[
		main <- (({[^A-Za-z0-9_]*} string)* ({[^A-Za-z0-9_]*})?) -> {}
		string <- ({[A-Za-z0-9_] [A-Za-z0-9_]*})
	]]
	
	local newString = ""
	local parsedTable = re.match(str, gramar)
	local replacementTable = replacementTable()
	for _, a in pairs(parsedTable) do 
		if replacementTable[string.lower(a)]~= nil then newString = newString .. replacementTable[string.lower(a)]
		else newString = newString .. a end
	end
	print(newString)
	print(dumptable(re.match(newString, sparqlExpressionGrammar())) , re.match(newString, sparqlExpressionGrammar()) or "NOT")
end

function replacementTable()
	local replacementTable = {
		["in"] = "IN",
		["not"] = "NOT",
		["str"] = "STR",
		["lang"] = "LANG",
		["langmatches"] = "LANGMATCHES",
		["datatype"] = "DATATYPE",
		["bound"] = "BOUND",
		["iri"] = "IRI",
		["uri"] = "URI",
		["bnode"] = "BNODE",
		["rand"] = "RAND",
		["abs"] = "ABS",
		["ceil"] = "CEIL",
		["floor"] = "FLOOR",
		["round"] = "ROUND",
		["concat"] = "CONCAT",
		["strlen"] = "STRLEN",
		["ucase"] = "UCASE",
		["lcase"] = "LCASE",
		["encode_for_uri"] = "ENCODE_FOR_URI",
		["contains"] = "CONTAINS",
		["strstarts"] = "STRSTARTS",
		["strends"] = "STRENDS",
		["strbefore"] = "STRBEFORE",
		["strafter"] = "STRAFTER",
		["year"] = "YEAR",
		["month"] = "MONTH",
		["day"] = "DAY",
		["hours"] = "HOURS",
		["minutes"] = "MINUTES",
		["seconds"] = "SECONDS",
		["timezone"] = "TIMEZONE",
		["tz"] = "TZ",
		["now"] = "NOW",
		["uuid"] = "UUID",
		["struuid"] = "STRUUID",
		["md5"] = "MD5",
		["sha1"] = "SHA1",
		["sha256"] = "SHA256",
		["sha384"] = "SHA384",
		["sha512"] = "SHA512",
		["coalesce"] = "COALESCE",
		["if"] = "IF",
		["strlang"] = "STRLANG",
		["strdt"] = "STRDT",
		["sameterm"] = "sameTerm",
		["isiri"] = "isIRI",
		["isuri"] = "isURI",
		["isblank"] = "isBLANK",
		["isliteral"] = "isLITERAL",
		["isnumeric"] = "isNUMERIC",
		["distinct"] = "DISTINCT",
		["true"] = "true",
		["false"] = "true",
		["regex"] = "REGEX",
		["substr"] = "SUBSTR",
		["replace"] = "REPLACE",
		["exists"] = "EXISTS",
		["count"] = "COUNT",
		["sum"] = "SUM",
		["min"] = "MIN",
		["max"] = "MAX",
		["avg"] = "AVG",
		["sample"] = "SAMPLE",
		["group_concat"] = "GROUP_CONCAT",
		["separator"] = "SEPARATOR",
		["or"] = "OR",
		["and"] = "AND",
	}
	return replacementTable
end