module(..., package.seeall)

require "core"

function setShowInstanceName(compartment)
	local elem = core.get_compartment_element(compartment)
	local name = elem:find("/compartment:has(/compartType[id='Name'])")
	local instanceName = compartment:attr("value")
end

function setInstanceNameOnClose(form)
	local instance = form:find("/component/component/component[id='Instance']/component[id='field']/compartment")
	local name = form:find("/component/component/component[id='Name']/component[id='field']/compartment"):attr("value")
	
	local toolType = lQuery("ToolType")
	if toolType:find("/tag[key = 'counter']"):is_empty() then lQuery.create("Tag", {key = 'counter', value = 0}):link("thing", toolType) end
	
	if instance:attr("value") == "" and name~="" then 
		instance:attr("value", name.."_"..tonumber(toolType:find("/tag[key = 'counter']"):attr("value"))+1)
		core.update_compartment_input_from_value(instance)
	
		toolType:find("/tag[key = 'counter']"):attr("value", tonumber(toolType:find("/tag[key = 'counter']"):attr("value"))+1)
	end
end

function setInstance(compartment)
	local elem = core.get_compartment_element(compartment)
	
	local name = compartment:attr("value")
	local instance = elem:find("/compartment:has(/compartType[id='Instance'])")
	
	local toolType = lQuery("ToolType")
	if toolType:find("/tag[key = 'counter']"):is_empty() then lQuery.create("Tag", {key = 'counter', value = 0}):link("thing", toolType) end
	
	if instance:attr("value") == "" and name~="" then 
		instance:attr("value", name.."_"..tonumber(toolType:find("/tag[key = 'counter']"):attr("value"))+1)
		core.update_compartment_input_from_value(compartment)
	
		instance:find("/component"):attr("text", instance:attr("value"))
		local cmd = utilities.create_command("D#Command", {info = "Refresh"})
		instance:find("/component"):link("command", cmd)
		utilities.execute_cmd_obj(cmd)
	
		toolType:find("/tag[key = 'counter']"):attr("value", tonumber(toolType:find("/tag[key = 'counter']"):attr("value"))+1)
	end
end

function parseSPARQL(comp)
    print(comp:find("/presentationElement"):attr("value"), comp:attr("id"))
	local sp = require "sparql_grammar"
	sp.parsesparqlExpressionGrammar(comp:find("/presentationElement"):attr("value"))
end

function setInstanceNameVisible(compartment)
	local element = compartment:find("/element")
	local instanceCompartment = element:find("/compartment:has(/compartType[id='Instance'])")
	instanceCompartment:remove_link("compartStyle", instanceCompartment:find("/elemStyle"))
	if compartment:attr("value") == "true" then 
		instanceCompartment:link("compartStyle", instanceCompartment:find("/compartType/compartStyle[id='Instance']"))
	else
		instanceCompartment:link("compartStyle", instanceCompartment:find("/compartType/compartStyle[id='InstanceNotVissible']"))
	end
end

function setEnabledFields(form)
	lQuery("Notation"):each(function(notation)
		if notation:attr("value")~= "" then
			notation:attr("value", string.gsub(notation:attr("value"), "<<", ""))
			notation:attr("value", string.gsub(notation:attr("value"), ">>", ""))
		end
	end)
	setEnabledField("OrderBy", form)
	setEnabledField("ASFictitiousHaving", form)
	setEnabledDistinct(form)
	setEnabledField("ShowRows", form)
	setEnabledField("SkipRows", form)
	-- setEnabledField("FullSPARQL", form)
end

function setEnabledField(fieldName, form)
	local orderBy = form:find("/component/component/component[id='" .. fieldName .. "']/component")
	local classType = form:find("/component/component/component[id='ClassType']/component[id='field']")
	if classType:attr("text") == "condition" then orderBy:attr("enabled", false) end
end

function setEnabledDistinct(form)
	local Distinct = form:find("/component/component/component[id='Distinct']/component/component")
	local classType = form:find("/component/component/component[id='ClassType']/component[id='field']")
	if classType:attr("text") == "condition" then Distinct:attr("enabled", false) end
end

function setEnabledFieldsFromClassTypeChange(compartment)
    local elem = core.get_compartment_element(compartment)
	local classType = elem:find("/compartment:has(/compartType[id='ClassType'])")
	local enabled = true
	if classType:attr("value") == "condition" then
	    enabled = false
	else
		enabled = true
	end
	setEnabledFieldFromClassTypeChange('ASFictitiousOrderBy', compartment, elem, enabled)
	setEnabledFieldFromClassTypeChange('ASFictitiousHaving', compartment, elem, enabled)
	setEnabledFieldFromClassTypeChange('ShowRows', compartment, elem, enabled)
	setEnabledFieldFromClassTypeChange('SkipRows', compartment, elem, enabled)
	setEnabledFieldFromClassTypeChange('Distinct', compartment, elem, enabled)
	-- setEnabledFieldFromClassTypeChange('FullSPARQL', compartment, elem, enabled)
end

function setEnabledFieldFromClassTypeChange(fieldName, compartment, elem, enabled)
	local field = elem:find("/compartment:has(/compartType[id='" .. fieldName .. "'])/component/container/component")
	field:attr("enabled", enabled)
	local cmd = utilities.create_command("D#Command", {info = "Refresh"})
	elem:find("/compartment:has(/compartType[id='" .. fieldName .. "'])/component/container"):link("command", cmd)
	utilities.execute_cmd_obj(cmd)
end

function setClassTypeValue(element)
	-- local sp = require "sparql_grammar"
	-- sp.parsesparqlExpressionGrammar()
	local classType = lQuery("Event/paletteBox"):attr("caption")
	if classType == "ConditionClass" then
		local classTypeCompartment = element:find("/compartment:has(/compartType[id='ClassType'])")
		classTypeCompartment:attr("value", "condition")
		-- classTypeCompartment:attr("input", "condition")
		
		element:remove_link("elemStyle", element:find("/elemStyle"))
		element:link("elemStyle", element:find("/elemType/elemStyle[id='ConditionClass']"))
	end
	
	--[[local graphDiagram = element:find("/graphDiagram")
	if(graphDiagram:find("/element/compartment[value='query']:has(/compartType[id='ClassType'])"):size() > 1) then
		local classTypeCompartment = element:find("/compartment:has(/compartType[id='ClassType'])")
		classTypeCompartment:attr("value", "condition")
		-- classTypeCompartment:attr("input", "condition")
		
		element:remove_link("elemStyle", element:find("/elemStyle"))
		element:link("elemStyle", element:find("/elemType/elemStyle[id='ConditionClass']"))
	end--]]
	
end

function setSubQueryInverseLink(compartment)
    local element = compartment:find("/element")
	local nameCompartment = element:find("/compartment:has(/compartType[id='Name'])")
	core.update_compartment_input_from_value(nameCompartment)
end

function setSubQueryNamePrefix(dataCompartType, dataCompartment, parsingCompartment)
	local result = ""
	local inverseLink = dataCompartment:find("/element/compartment:has(/compartType[id='IsInverse'])")
	if inverseLink:attr("value") == "true" then result = "inv(" else result = "" end
	return result
end

function setSubQueryNameSuffix(dataCompartType, dataCompartment, parsingCompartment)
	local result = ""
	local inverseLink = dataCompartment:find("/element/compartment:has(/compartType[id='IsInverse'])")
	if inverseLink:attr("value") == "true" then result = ")" else result = "" end
	return result
end

function setIsInverse(compartment)
	local elem = core.get_compartment_element(compartment)
	local className2 = elem:find("/end/compartment:has(/compartType[id='Name'])"):attr("value")
	local class2 = lQuery("Onto#Class[name='" .. className2 .. "']")
	
	local needToSetIsInverse = false
	
	class2:find("/domainAssoc"):filter("Onto#Association,Onto#SuperAssociation"):each(
	function (assoc)
		if assoc:attr("name") == compartment:attr("value") then needToSetIsInverse = true end
	end)
	
	local isInverse = elem:find("/compartment:has(/compartType[id='IsInverse'])")
	isInverse:attr("value", needToSetIsInverse)
	
	isInverse:find("/component"):attr("checked", needToSetIsInverse)
	
	core.update_compartment_input_from_value(compartment)
	
	local cmd = utilities.create_command("D#Command", {info = "Refresh"})
	isInverse:find("/component"):link("command", cmd)
	utilities.execute_cmd_obj(cmd)
end

function setClassTypeFromSubQuery(element)
	local subQueryElement = element:find("/end")
	
	local classTypeCompartment = subQueryElement:find("/compartment:has(/compartType[id='ClassType'])")
	classTypeCompartment:attr("value", "subquery")
	-- classTypeCompartment:attr("input", "subquery")
	
	subQueryElement:remove_link("elemStyle", subQueryElement:find("/elemStyle"))
	subQueryElement:link("elemStyle", subQueryElement:find("/elemType/elemStyle[id='SubQueryClass']"))
	
	utilities.refresh_element(subQueryElement, subQueryElement:find("/graphDiagram")) 
	utilities.execute_cmd("SaveDgrCmd", {graphDiagram = diagram})
end

function setIsOptional(compartment)
	
	local com = compartment:filter(function(obj)
		return obj:find("/compartType"):attr("id") == "IsOptional"
	end)
	
	if com:attr("value") == "true" then
		local element = com:find("/element")
		local isNegationCompartment = element:find("/compartment:has(/compartType[id='IsNegation'])")
		isNegationCompartment:attr("value", "")
		core.update_compartment_input_from_value(isNegationCompartment)
		
		isNegationCompartment:find("/component"):attr("checked", "false")
		
		local cmd = utilities.create_command("D#Command", {info = "Refresh"})
		isNegationCompartment:find("/component"):link("command", cmd)
		utilities.execute_cmd_obj(cmd)
		
		element:remove_link("elemStyle", element:find("/elemStyle"))
		element:link("elemStyle", element:find("/elemType/elemStyle[id='Optional Link']"))
		
		utilities.refresh_element(element, element:find("/graphDiagram")) 
		utilities.execute_cmd("SaveDgrCmd", {graphDiagram = diagram})
	end
end

function setIsNegation(compartment)	
	local com = compartment:filter(function(obj)
		return obj:find("/compartType"):attr("id") == "IsNegation"
	end)

	if com:attr("value") == "true" then
		local element = com:find("/element")
		local isNegationCompartment = element:find("/compartment:has(/compartType[id='IsOptional'])")
		isNegationCompartment:attr("value", "")
		core.update_compartment_input_from_value(isNegationCompartment)
		
		isNegationCompartment:find("/component"):attr("checked", "false")
		
		local cmd = utilities.create_command("D#Command", {info = "Refresh"})
		isNegationCompartment:find("/component"):link("command", cmd)
		utilities.execute_cmd_obj(cmd)
		
		element:remove_link("elemStyle", element:find("/elemStyle"))
		element:link("elemStyle", element:find("/elemType/elemStyle[id='NegationLink']"))
		
		utilities.refresh_element(element, element:find("/graphDiagram")) 
		utilities.execute_cmd("SaveDgrCmd", {graphDiagram = diagram})
	end
end

function setIsCondition(compartment)	
	setIsGroupIsCondition(compartment, "IsCondition", "IsGroup")
end

function setIsGroup(compartment)	
	setIsGroupIsCondition(compartment, "IsGroup", "IsCondition")
end

function setIsGroupIsCondition(compartment, compartmentName, oppositeCompartmentName)
	local com = compartment:filter(function(obj)
		return obj:find("/compartType"):attr("id") == compartmentName
	end)

	if com:attr("value") == "true" then
		local element = com:find("/element")
		local isGroupCompartment = element:find("/compartment:has(/compartType[id='" .. oppositeCompartmentName .. "'])")
		isGroupCompartment:attr("value", "")
		core.update_compartment_input_from_value(isGroupCompartment)
		
		isGroupCompartment:find("/component"):attr("checked", "false")
		
		local cmd = utilities.create_command("D#Command", {info = "Refresh"})
		isGroupCompartment:find("/component"):link("command", cmd)
		utilities.execute_cmd_obj(cmd)
	end
end

function generateDate(compartment)
	local elem = utilities.active_elements()
	local compart = elem:find("/compartment:has(/compartType[id = 'Name'])")
	local target_compart = compart:find("/element/compartment:has(/compartType[id = 'Date'])")
	core.set_compartment_and_field(target_compart,  os.date("%Y-%m-%dT%H:%M:%S"))
	utilities.update_target_diagram_caption(compartment)
end

function attributeGrammar()  
   local grammar = [[( {:Show: 'define ' :} )?
{('')}? ( {:IsOptional: 'optional ' :} )?
{('')}? ( {:Alias: [a-zA-Z0-9-_]* :} '=')?
{('')}? ( {:AttrName: [][a-zA-Z0-9-_!?@#$%^&*()=*+"':;,./ {}<>]* :} )? !.]]

local grammar2 = [[( {:Show: 'define ' :} )?
{('')}? ( {:IsOptional: 'optional ' :} )?
{('')}? ( {:Alias: [a-zA-Z0-9-_]* :} '=')?
{('')}? ( {:Aggregation: ('count' / 'count_distinct' / 'sum' / 'max' / 'min' / 'avg' / 'group_concat' / 'sample') ' '? :})?
(({('')}? ( {:AttrName: [a-zA-Z0-9-_:]* :} )? !.) /
({('')}? ({:Expression: [][a-zA-Z0-9-_!?@#$%^&*()=*+"':;,./ {}<>]* :} )? !.))]]
return grammar
end

function setShowValue(attributesForm)
	local attribute = attributesForm:find("/presentationElement")
	local show = attribute:find("/subCompartment:has(/compartType[id='Show'])")
	local value
	if string.sub(attribute:attr("value"),1,7)=="define " then value = false else value = true end

	show:attr("value", value)
	local checkBox = show:find("/component")
	if checkBox:is_not_empty() then 
		checkBox:attr("checked", value)
	else attributesForm:find("/component[id='Show']/component/component[id='field']"):attr("checked", value) end
end

function getClassNames(compart)
	local klases = {}
	--create table of all classes
	lQuery("Onto#Class"):each(
		function (class)
			table.insert(klases, class:attr("name"))
		end
	)
	--sort table according to alphabet
	table.sort(klases, function (a, b)
		-- return string.lower(a.name) < string.lower(b.name)
		return string.lower(a) < string.lower(b)
		end
	)
	return klases
end

function getAttributeNames(compart)
	local klases = {}
	local elem = core.get_compartment_element(compart)
	local name = elem:find("/compartment:has(/compartType[id = 'Name'])"):attr_e("value")
	if name ~= "" then
		local class = lQuery("Onto#Class[name='" .. name .. "']")
		class:transitive_closure("/superClass"):find("/attribute"):each(
			function (attribute)
				table.insert(klases, attribute:attr("name"))
			end
		)
		--sort table according to alphabet
		table.sort(klases, function (a, b)
			return string.lower(a) < string.lower(b)
			end
		)
	end
	--agregation patterns
	local agreggationTable = {"count", "count distinct", "sum", "max", "min", "avg", "group_concat", "sample"}
	for _, agreggation in pairs(agreggationTable) do
		table.insert(klases, agreggation .. "(...)")
	end
	return klases
end

function getAssociationNames(compart)
	local klases = {}
	local elem = core.get_compartment_element(compart)
	local className1 = elem:find("/start/compartment:has(/compartType[id='Name'])"):attr("value")
	local className2 = elem:find("/end/compartment:has(/compartType[id='Name'])"):attr("value")
	local class1 = lQuery("Onto#Class[name='" .. className1 .. "']")
	local class2 = lQuery("Onto#Class[name='" .. className2 .. "']")
	class1:find("/domainAssoc"):filter("Onto#Association,Onto#SuperAssociation"):each(
	function (assoc)
		if assoc:find("/range"):attr("name") == className2 then table.insert(klases, assoc:attr("name")) end
	end)
	class2:find("/domainAssoc"):filter("Onto#Association,Onto#SuperAssociation"):each(
	function (assoc)
		if assoc:find("/range"):attr("name") == className1 then table.insert(klases, assoc:attr("name")) end
	end)
	return klases
end